from rest_framework import status
from rest_framework.response import Response

from accu.core.helpers.clean import remove_non_printable_characters, strip_html_tags


class CleanMixin:
    """Model mixin class which cleans inputs using the Mozilla bleach tools."""

    # Define a list of field names which will *not* be cleaned
    SAFE_FIELDS = []

    def create(self, request, *args, **kwargs):
        """Override to clean data before processing it."""
        serializer = self.get_serializer(data=self.clean_data(request.data))
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def update(self, request, *args, **kwargs):
        """Override to clean data before processing it."""
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(
            instance, data=self.clean_data(request.data), partial=partial
        )
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response(serializer.data)

    def clean_string(self, field: str, data: str) -> str:
        """Clean / sanitize a single input string.

        Note that this function will *allow* orphaned <>& characters,
        which would normally be escaped by bleach.

        Nominally, the only thing that will be "cleaned" will be HTML tags

        Ref: https://github.com/mozilla/bleach/issues/192
        """

        cleaned = strip_html_tags(data, field_name=field)
        cleaned = remove_non_printable_characters(cleaned)

        return cleaned

    def clean_data(self, data: dict) -> dict:
        """Clean / sanitize data.

        This uses mozillas bleach under the hood to disable certain html tags by
        encoding them - this leads to script tags etc. to not work.
        The results can be longer then the input; might make some character combinations
        `ugly`. Prevents XSS on the server-level.

        Args:
            data (dict): Data that should be sanatized.

        Returns:
            dict: Provided data sanatized; still in the same order.
        """

        clean_data = {}

        for k, v in data.items():

            if k in self.SAFE_FIELDS:
                ret = v
            elif isinstance(v, str):
                ret = self.clean_string(k, v)
            elif isinstance(v, dict):
                ret = self.clean_data(v)
            else:
                ret = v

            clean_data[k] = ret

        return clean_data



from .svg import ALLOWED_ELEMENTS_SVG, sanitize_svg

SVG_BAS_STRING = "<svg><script>alert('xss')</script></svg>"


def test_true():
    assert True


def test_sanitize_svg():
    """Test that the svg sanitizer works."""
    assert sanitize_svg(SVG_BAS_STRING) == "<svg>alert('xss')</svg>"
    assert (
        sanitize_svg(SVG_BAS_STRING, strip=False)
        == "<svg>&lt;script&gt;alert('xss')&lt;/script&gt;</svg>"
    )
    assert sanitize_svg(SVG_BAS_STRING, strip=True) == "<svg>alert('xss')</svg>"

    ### Advanced options
    EXTENDED_BASE_STRING = f"<xyz>{SVG_BAS_STRING}</xyz>"
    # Normal run
    assert sanitize_svg(EXTENDED_BASE_STRING) == "<svg>alert('xss')</svg>"
    # Allow custom tag
    assert (
        sanitize_svg(EXTENDED_BASE_STRING, elements=ALLOWED_ELEMENTS_SVG + ["xyz"])
        == "<xyz><svg>alert('xss')</svg></xyz>"
    )



from functools import wraps


def auth_exempt(view_func):
    """Mark a view function as being exempt from auth requirements."""

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.auth_exempt = True
    return wraps(view_func)(wrapped_view)



from .validators import allowable_url_schemes


def test_true():
    assert True


def test_allowable_url_schemes():
    """Test the allowable_url_schemes function."""
    assert allowable_url_schemes() == ["http", "https", "ftp", "ftps", "sftp"]
    assert ["weddav"] in allowable_url_schemes("weddav")



import regex
from bleach import clean
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def strip_html_tags(value: str, raise_error=True, field_name=None):
    """Strip HTML tags from an input string using the bleach library.

    If raise_error is True, a ValidationError will be thrown if HTML tags are detected
    """

    cleaned = clean(
        value,
        strip=True,
        tags=[],
        attributes=[],
    )

    # Add escaped characters back in
    replacements = {
        "&gt;": ">",
        "&lt;": "<",
        "&amp;": "&",
    }

    for o, r in replacements.items():
        cleaned = cleaned.replace(o, r)

    # If the length changed, it means that HTML tags were removed!
    if len(cleaned) != len(value) and raise_error:

        field = field_name or "non_field_errors"

        raise ValidationError({field: [_("Remove HTML tags from this value")]})

    return cleaned


def remove_non_printable_characters(value: str, remove_ascii=True, remove_unicode=True):
    """Remove non-printable / control characters from the provided string"""

    if remove_ascii:
        # Remove ASCII control characters
        cleaned = regex.sub("[\x01-\x1F]+", "", value)

    if remove_unicode:
        # Remove Unicode control characters
        cleaned = regex.sub(r"[^\P{C}]+", "", value)

    return cleaned



from pathlib import Path


def is_true(x):
    """Shortcut function to determine if a value "looks" like a boolean"""
    return str(x).strip().lower() in ["1", "y", "yes", "t", "true", "on"]


def get_base_dir() -> Path:
    """Returns the base (top-level) InvenTree directory."""
    return Path(__file__).parent.parent.resolve()



"""Functions to sanitize user input files."""
from bleach import clean
from bleach.css_sanitizer import CSSSanitizer

ALLOWED_ELEMENTS_SVG = [
    "a",
    "animate",
    "animateColor",
    "animateMotion",
    "animateTransform",
    "circle",
    "defs",
    "desc",
    "ellipse",
    "font-face",
    "font-face-name",
    "font-face-src",
    "g",
    "glyph",
    "hkern",
    "linearGradient",
    "line",
    "marker",
    "metadata",
    "missing-glyph",
    "mpath",
    "path",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "set",
    "stop",
    "svg",
    "switch",
    "text",
    "title",
    "tspan",
    "use",
]

ALLOWED_ATTRIBUTES_SVG = [
    "accent-height",
    "accumulate",
    "additive",
    "alphabetic",
    "arabic-form",
    "ascent",
    "attributeName",
    "attributeType",
    "baseProfile",
    "bbox",
    "begin",
    "by",
    "calcMode",
    "cap-height",
    "class",
    "color",
    "color-rendering",
    "content",
    "cx",
    "cy",
    "d",
    "dx",
    "dy",
    "descent",
    "display",
    "dur",
    "end",
    "fill",
    "fill-opacity",
    "fill-rule",
    "font-family",
    "font-size",
    "font-stretch",
    "font-style",
    "font-variant",
    "font-weight",
    "from",
    "fx",
    "fy",
    "g1",
    "g2",
    "glyph-name",
    "gradientUnits",
    "hanging",
    "height",
    "horiz-adv-x",
    "horiz-origin-x",
    "id",
    "ideographic",
    "k",
    "keyPoints",
    "keySplines",
    "keyTimes",
    "lang",
    "marker-end",
    "marker-mid",
    "marker-start",
    "markerHeight",
    "markerUnits",
    "markerWidth",
    "mathematical",
    "max",
    "min",
    "name",
    "offset",
    "opacity",
    "orient",
    "origin",
    "overline-position",
    "overline-thickness",
    "panose-1",
    "path",
    "pathLength",
    "points",
    "preserveAspectRatio",
    "r",
    "refX",
    "refY",
    "repeatCount",
    "repeatDur",
    "requiredExtensions",
    "requiredFeatures",
    "restart",
    "rotate",
    "rx",
    "ry",
    "slope",
    "stemh",
    "stemv",
    "stop-color",
    "stop-opacity",
    "strikethrough-position",
    "strikethrough-thickness",
    "stroke",
    "stroke-dasharray",
    "stroke-dashoffset",
    "stroke-linecap",
    "stroke-linejoin",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-width",
    "systemLanguage",
    "target",
    "text-anchor",
    "to",
    "transform",
    "type",
    "u1",
    "u2",
    "underline-position",
    "underline-thickness",
    "unicode",
    "unicode-range",
    "units-per-em",
    "values",
    "version",
    "viewBox",
    "visibility",
    "width",
    "widths",
    "x",
    "x-height",
    "x1",
    "x2",
    "xlink:actuate",
    "xlink:arcrole",
    "xlink:href",
    "xlink:role",
    "xlink:show",
    "xlink:title",
    "xlink:type",
    "xml:base",
    "xml:lang",
    "xml:space",
    "xmlns",
    "xmlns:xlink",
    "y",
    "y1",
    "y2",
    "zoomAndPan",
    "style",
]


def sanitize_svg(
    file_data: str,
    strip: bool = True,
    elements: str = ALLOWED_ELEMENTS_SVG,
    attributes: str = ALLOWED_ATTRIBUTES_SVG,
) -> str:
    """Sanatize a SVG file.

    Args:
        file_data (str): SVG as string.
        strip (bool, optional): Should invalid elements get removed. Defaults to True.
        elements (str, optional): Allowed elements. Defaults to ALLOWED_ELEMENTS_SVG.
        attributes (str, optional): Allowed attributes. Defaults to ALLOWED_ATTRIBUTES_SVG.

    Returns:
        str: Sanitzied SVG file.
    """

    cleaned = clean(
        file_data,
        tags=elements,
        attributes=attributes,
        strip=strip,
        strip_comments=strip,
        css_sanitizer=CSSSanitizer(),
    )
    return cleaned



from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt


class CsrfExemptMixin(object):
    """Exempts the view from CSRF requirements."""

    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        """Overwrites dispatch to be extempt from csrf checks."""
        return super().dispatch(*args, **kwargs)
